# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class PipelineStepCreateAPIInput(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'description': 'str',
        'external_service_spec_input': 'V1ExternalServiceSpecInput',
        'fail_spec_input': 'object',
        'if_spec_input': 'V1IfSpecInput',
        'manual_input_spec_input': 'V1ManualInputSpecInput',
        'manual_judgment_spec_input': 'V1ManualJudgmentSpecInput',
        'notification_spec_input': 'V1NotificationSpecInput',
        'run_spec_input': 'V1RunSpecInput',
        'step_key': 'str',
        'step_type': 'str',
        'title': 'str',
        'trigger_dispatch_input': 'V1TriggerDispatchInput'
    }

    attribute_map = {
        'description': 'description',
        'external_service_spec_input': 'external_service_spec_input',
        'fail_spec_input': 'fail_spec_input',
        'if_spec_input': 'if_spec_input',
        'manual_input_spec_input': 'manual_input_spec_input',
        'manual_judgment_spec_input': 'manual_judgment_spec_input',
        'notification_spec_input': 'notification_spec_input',
        'run_spec_input': 'run_spec_input',
        'step_key': 'step_key',
        'step_type': 'step_type',
        'title': 'title',
        'trigger_dispatch_input': 'trigger_dispatch_input'
    }

    def __init__(self, description=None, external_service_spec_input=None, fail_spec_input=None, if_spec_input=None, manual_input_spec_input=None, manual_judgment_spec_input=None, notification_spec_input=None, run_spec_input=None, step_key=None, step_type=None, title=None, trigger_dispatch_input=None, local_vars_configuration=None):  # noqa: E501
        """PipelineStepCreateAPIInput - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._description = None
        self._external_service_spec_input = None
        self._fail_spec_input = None
        self._if_spec_input = None
        self._manual_input_spec_input = None
        self._manual_judgment_spec_input = None
        self._notification_spec_input = None
        self._run_spec_input = None
        self._step_key = None
        self._step_type = None
        self._title = None
        self._trigger_dispatch_input = None
        self.discriminator = None

        self.description = description
        if external_service_spec_input is not None:
            self.external_service_spec_input = external_service_spec_input
        if fail_spec_input is not None:
            self.fail_spec_input = fail_spec_input
        if if_spec_input is not None:
            self.if_spec_input = if_spec_input
        if manual_input_spec_input is not None:
            self.manual_input_spec_input = manual_input_spec_input
        if manual_judgment_spec_input is not None:
            self.manual_judgment_spec_input = manual_judgment_spec_input
        if notification_spec_input is not None:
            self.notification_spec_input = notification_spec_input
        if run_spec_input is not None:
            self.run_spec_input = run_spec_input
        self.step_key = step_key
        self.step_type = step_type
        self.title = title
        if trigger_dispatch_input is not None:
            self.trigger_dispatch_input = trigger_dispatch_input

    @property
    def description(self):
        """Gets the description of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The description of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this PipelineStepCreateAPIInput.


        :param description: The description of this PipelineStepCreateAPIInput.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def external_service_spec_input(self):
        """Gets the external_service_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The external_service_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1ExternalServiceSpecInput
        """
        return self._external_service_spec_input

    @external_service_spec_input.setter
    def external_service_spec_input(self, external_service_spec_input):
        """Sets the external_service_spec_input of this PipelineStepCreateAPIInput.


        :param external_service_spec_input: The external_service_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type external_service_spec_input: V1ExternalServiceSpecInput
        """

        self._external_service_spec_input = external_service_spec_input

    @property
    def fail_spec_input(self):
        """Gets the fail_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The fail_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: object
        """
        return self._fail_spec_input

    @fail_spec_input.setter
    def fail_spec_input(self, fail_spec_input):
        """Sets the fail_spec_input of this PipelineStepCreateAPIInput.


        :param fail_spec_input: The fail_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type fail_spec_input: object
        """

        self._fail_spec_input = fail_spec_input

    @property
    def if_spec_input(self):
        """Gets the if_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The if_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1IfSpecInput
        """
        return self._if_spec_input

    @if_spec_input.setter
    def if_spec_input(self, if_spec_input):
        """Sets the if_spec_input of this PipelineStepCreateAPIInput.


        :param if_spec_input: The if_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type if_spec_input: V1IfSpecInput
        """

        self._if_spec_input = if_spec_input

    @property
    def manual_input_spec_input(self):
        """Gets the manual_input_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The manual_input_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1ManualInputSpecInput
        """
        return self._manual_input_spec_input

    @manual_input_spec_input.setter
    def manual_input_spec_input(self, manual_input_spec_input):
        """Sets the manual_input_spec_input of this PipelineStepCreateAPIInput.


        :param manual_input_spec_input: The manual_input_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type manual_input_spec_input: V1ManualInputSpecInput
        """

        self._manual_input_spec_input = manual_input_spec_input

    @property
    def manual_judgment_spec_input(self):
        """Gets the manual_judgment_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The manual_judgment_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1ManualJudgmentSpecInput
        """
        return self._manual_judgment_spec_input

    @manual_judgment_spec_input.setter
    def manual_judgment_spec_input(self, manual_judgment_spec_input):
        """Sets the manual_judgment_spec_input of this PipelineStepCreateAPIInput.


        :param manual_judgment_spec_input: The manual_judgment_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type manual_judgment_spec_input: V1ManualJudgmentSpecInput
        """

        self._manual_judgment_spec_input = manual_judgment_spec_input

    @property
    def notification_spec_input(self):
        """Gets the notification_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The notification_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1NotificationSpecInput
        """
        return self._notification_spec_input

    @notification_spec_input.setter
    def notification_spec_input(self, notification_spec_input):
        """Sets the notification_spec_input of this PipelineStepCreateAPIInput.


        :param notification_spec_input: The notification_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type notification_spec_input: V1NotificationSpecInput
        """

        self._notification_spec_input = notification_spec_input

    @property
    def run_spec_input(self):
        """Gets the run_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The run_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1RunSpecInput
        """
        return self._run_spec_input

    @run_spec_input.setter
    def run_spec_input(self, run_spec_input):
        """Sets the run_spec_input of this PipelineStepCreateAPIInput.


        :param run_spec_input: The run_spec_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type run_spec_input: V1RunSpecInput
        """

        self._run_spec_input = run_spec_input

    @property
    def step_key(self):
        """Gets the step_key of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The step_key of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._step_key

    @step_key.setter
    def step_key(self, step_key):
        """Sets the step_key of this PipelineStepCreateAPIInput.


        :param step_key: The step_key of this PipelineStepCreateAPIInput.  # noqa: E501
        :type step_key: str
        """
        if self.local_vars_configuration.client_side_validation and step_key is None:  # noqa: E501
            raise ValueError("Invalid value for `step_key`, must not be `None`")  # noqa: E501

        self._step_key = step_key

    @property
    def step_type(self):
        """Gets the step_type of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The step_type of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._step_type

    @step_type.setter
    def step_type(self, step_type):
        """Sets the step_type of this PipelineStepCreateAPIInput.


        :param step_type: The step_type of this PipelineStepCreateAPIInput.  # noqa: E501
        :type step_type: str
        """
        if self.local_vars_configuration.client_side_validation and step_type is None:  # noqa: E501
            raise ValueError("Invalid value for `step_type`, must not be `None`")  # noqa: E501
        allowed_values = ["v1/run", "v1/fail", "v1/if", "v1/manual_input", "v1/manual_judgment", "v1/external_service", "v1/notification", "v1/trigger_dispatch"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and step_type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `step_type` ({0}), must be one of {1}"  # noqa: E501
                .format(step_type, allowed_values)
            )

        self._step_type = step_type

    @property
    def title(self):
        """Gets the title of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The title of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this PipelineStepCreateAPIInput.


        :param title: The title of this PipelineStepCreateAPIInput.  # noqa: E501
        :type title: str
        """
        if self.local_vars_configuration.client_side_validation and title is None:  # noqa: E501
            raise ValueError("Invalid value for `title`, must not be `None`")  # noqa: E501

        self._title = title

    @property
    def trigger_dispatch_input(self):
        """Gets the trigger_dispatch_input of this PipelineStepCreateAPIInput.  # noqa: E501


        :return: The trigger_dispatch_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :rtype: V1TriggerDispatchInput
        """
        return self._trigger_dispatch_input

    @trigger_dispatch_input.setter
    def trigger_dispatch_input(self, trigger_dispatch_input):
        """Sets the trigger_dispatch_input of this PipelineStepCreateAPIInput.


        :param trigger_dispatch_input: The trigger_dispatch_input of this PipelineStepCreateAPIInput.  # noqa: E501
        :type trigger_dispatch_input: V1TriggerDispatchInput
        """

        self._trigger_dispatch_input = trigger_dispatch_input

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PipelineStepCreateAPIInput):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, PipelineStepCreateAPIInput):
            return True

        return self.to_dict() != other.to_dict()



"""
Serving command definition and delegation.
"""

import json
import sys
from typing import List, Optional, TextIO

import click

from vessl.openapi_client import SecretGenericSecretUpsertAPIInput
from vessl.openapi_client.models import (
    ModelserviceModelServiceRevisionListResponse,
    ResponseModelServiceInfo,
    ResponseModelServiceRevision,
    secret_generic_secret_upsert_api_input,
)
from vessl import vessl_api
from vessl.cli._base import VesslCommand, VesslGroup, vessl_argument, vessl_option
from vessl.cli.serve.util import (
    print_gateway,
    print_revision,
    print_revisions,
    print_service,
)
from vessl.enums import ModelServiceType
from vessl.serving import _read_service, terminate_revision
from vessl.util.deprecaate import deprecated
from vessl.util.echo import print_info
from vessl.util.exception import VesslApiException
from vessl.util.prompt import prompt_choices, prompt_text

from .command_options import (
    autoscaling_max_option,
    autoscaling_metric_option,
    autoscaling_min_option,
    autoscaling_target_option,
    detail_option,
    file_path_option,
    force_option,
    format_option,
    hub_key_option,
    no_prompt_option,
    revision_option,
    serverless_service_option,
    service_activate_option,
    service_launch_option,
    service_name_option,
    service_name_or_new_option,
    split_traffic_revs_option,
    split_traffic_weights_option,
    nodeport_option, subdomain_option,
)
from .service import (
    abort_update,
    create,
    create_yaml,
    list,
    scale,
    set_access_token,
    split,
)

cli = VesslGroup("service")
alias_cli = VesslGroup("serve", help="Deprecated: Please use 'service'")


@alias_cli.group("revision")
def cli_revision():
    """
    (Deprecated) Root command for revision-related commands.
    """
    pass


@alias_cli.group("gateway")
def cli_gateway():
    """
    (Deprecated) Root command for gateway-related commands.
    """
    pass


@cli.command("read", cls=VesslCommand)
@service_name_option
@detail_option
@format_option
def revision_show(service: ResponseModelServiceInfo, detail: bool, format: str):
    """
    Show current status and information about a service.
    """
    print_service(service=service)

    try:
        resp: ModelserviceModelServiceRevisionListResponse = vessl_api.model_service_revision_list_api(
        organization_name=vessl_api.organization.name,
        model_service_name=service.name,
        )
        revisions = resp.results
    except VesslApiException as e:
        print(f"Failed to read service (name: {service.name}): {e.message}")
        sys.exit(1)

    if detail:
        full_revisions = [
            vessl_api.model_service_revision_read_api(
            organization_name=vessl_api.organization.name,
            model_service_name=service.name,
            revision_number=r.number,
        ) for r in revisions
        ]

    if format == "text":
        print_revisions(revisions=full_revisions if detail else revisions, verbose=detail)
    if format == "json":
        print(json.dumps([r.to_dict() for r in revisions], default=str))
    if format == "yaml":
        print("\n\n".join([r.yaml_spec for r in revisions]))


@cli_revision.command("terminate", cls=VesslCommand, hidden=True)
@service_name_option
@revision_option
@deprecated(target_command="serve revision terminate", new_command="service terminate")
def revision_terminate_deprecated(service: ResponseModelServiceInfo, number: ResponseModelServiceRevision):
    revision_terminate(service=service, revision=number)


@cli.command("terminate", cls=VesslCommand)
@service_name_option
@revision_option
def serve_terminate(service: ResponseModelServiceInfo, number: ResponseModelServiceInfo):
    """
    Terminate specified revision.
    """
    revision_terminate(service=service, revision=number)


def revision_terminate(service: ResponseModelServiceInfo, revision: ResponseModelServiceRevision):
    if revision.status in ["failed", "stopped"]:
        print("Selected revision is already terminated.")
        return

    try:
        if service.type == ModelServiceType.SERVERLESS.value:
            vessl_api.model_service_serverless_terminate_api(
                model_service_name=service.name,
                organization_name=vessl_api.organization.name
            )
        else:
            terminate_revision(
                organization=vessl_api.organization.name,
                service_name=service.name,
                revision_number=revision.number,
            )
        print("Successfully terminated revision.")
    except VesslApiException as e:
        print(f"Failed to terminate revision #{revision.number} of service {service.name}: {e.message}")
        sys.exit(1)


@cli.command("list", cls=VesslCommand)
def service_list():
    """
    List services in current organization.
    """
    list()


@alias_cli.command("list", cls=VesslCommand, hidden=True)
def serve_list(): list()


@cli.command("abort-update", cls=VesslCommand)
@service_name_option
def service_abort_update(service: ResponseModelServiceInfo):
    """
    Abort the current revision update.
    """
    abort_update(service=service)


@alias_cli.command("abort-update", cls=VesslCommand, hidden=True)
@service_name_option
def serve_abort_update(service: ResponseModelServiceInfo):
    abort_update(service=service)


@cli.command("create", cls=VesslCommand)
@hub_key_option
@file_path_option
@no_prompt_option
@service_launch_option
@service_activate_option
@force_option
@service_name_or_new_option
@serverless_service_option
@nodeport_option
@subdomain_option
def revision_create_with_yaml(
        file: TextIO,
        no_prompt: bool,
        launch: bool,
        set_current_active: bool,
        force: bool,
        service_name: str,
        serverless: bool,
        from_hub: str,
        port: int,
        subdomain: str
):
    """
    Create a revision from a YAML file.

    The YAML file should contain the definition of the revision. See https://docs.vessl.ai/guides/serve/service-yaml

    Example:
    $ vessl service create -f service.yaml --set-current-active
    $ vessl service create --from-hub="vllm-service"
    """
    create(
        file=file, no_prompt=no_prompt, launch=launch,
        set_current_active=set_current_active, force=force,
        service_name=service_name, serverless=serverless,
        from_hub=from_hub, node_port=port, subdomain=subdomain
    )


@alias_cli.command("create", cls=VesslCommand, hidden=True)
@hub_key_option
@file_path_option
@no_prompt_option
@service_launch_option
@service_activate_option
@force_option
@service_name_or_new_option
@serverless_service_option
@nodeport_option
@subdomain_option
@deprecated(target_command="serve create", new_command="service create")
def revision_create_with_yaml(
        file: TextIO,
        no_prompt: bool,
        launch: bool,
        set_current_active: bool,
        force: bool,
        service_name: str,
        serverless: bool,
        from_hub: str,
        port: int,
        subdomain: str
):
    create(
        file=file, no_prompt=no_prompt, launch=launch,
        set_current_active=set_current_active, force=force,
        service_name=service_name, serverless=serverless,
        from_hub=from_hub, node_port=port, subdomain=subdomain
    )


@cli.command("scale", cls=VesslCommand)
@service_name_option
@revision_option
@autoscaling_min_option
@autoscaling_max_option
@autoscaling_target_option
@autoscaling_metric_option
def service_scale(
        service: ResponseModelServiceInfo,
        number: ResponseModelServiceRevision,
        min: Optional[int],
        max: Optional[int],
        target: Optional[int],
        metric: Optional[str],
):
    """
    Update revision's autoscaler config. Partial updates of autoscaling options (min, max, target, metric) allowed.
    If current revision does not have autoscaling set, it will default to (min=1, max=1, target=60, metric=cpu).

    Example:
    
    $ vessl service scale --service=my-service --number=1 --min=1
    $ vessl service scale --min=1 --max=5 --metric=nvidia.com/gpu
    """
    scale(service=service, number=number, min=min, max=max, target=target, metric=metric)


@cli.command("set-access-token", cls=VesslCommand)
@service_name_option
@click.option("--secret", type=str, help="Secret to set as access token.")
def service_set_access_token(
        service: ResponseModelServiceInfo,
        secret: Optional[str],
):
    """
    Set access token for the service.

    Example:

    $ vessl serve set-access-token --service=my-service
    $ vessl serve set-access-token --service=my-service --secret=my-secret
    """

    if not secret:
        opt = prompt_choices(
            "How do you want to set access token?",
            ["By linking an existing secret", "By creating a new secret and linking it"],
            "By linking an existing secret",
        )
        if opt == "By linking an existing secret":
            secrets = (vessl_api.
                       secret_list_api(organization_name=vessl_api.organization.name, kind="generic-secret").secrets)
            secret_name = prompt_choices(
                "Select a secret to link as the access token",
                [s.credentials_name for s in secrets],
            )
            secret_obj = next(s for s in secrets if s.credentials_name == secret_name)
        else:
            secret_name = prompt_text("Enter the name of the secret to create as the access token")
            secret_text = prompt_text("Enter the secret text")
            secret_obj = vessl_api.secret_generic_secret_upsert_api(
                organization_name=vessl_api.organization.name,
                secret_generic_secret_upsert_api_input=SecretGenericSecretUpsertAPIInput(
                    secret_name=secret_name,
                    value=secret_text,
                ),
            )
    else:
        secret_name = secret
        secrets = (vessl_api.
                   secret_list_api(organization_name=vessl_api.organization.name, kind="generic-secret").secrets)
        secret_obj = next(s for s in secrets if s.credentials_name == secret_name)

    set_access_token(service=service, secret=secret_obj)
    print_info(f"Secret {secret_obj.credentials_name} is now linked as an access token of service {service.name}.")
    print_info(f"Please include the following header in your requests: \"Authorization: Bearer (your secret goes here)\"")
    print_info(f"$ curl -XPOST -H 'Authorization: Bearer some-secret-text' https://your-service-endpoint.com")


@cli_revision.command("scale", cls=VesslCommand, hidden=True)
@service_name_option
@revision_option
@autoscaling_min_option
@autoscaling_max_option
@autoscaling_target_option
@autoscaling_metric_option
@deprecated(target_command="serve revision scale", new_command="service scale")
def serve_revision_scale(
        service: ResponseModelServiceInfo,
        number: ResponseModelServiceRevision,
        min: Optional[int],
        max: Optional[int],
        target: Optional[int],
        metric: Optional[str],
): scale(service=service, number=number, min=min, max=max, target=target, metric=metric)


@alias_cli.command("update", cls=VesslCommand, hidden=True)
@service_name_option
@split_traffic_revs_option
@split_traffic_weights_option
@click.option(
    "--interactive",
    is_flag=True,
    type=click.BOOL,
    help="Update service interactively"
)
@deprecated(target_command="serve update", new_command="service split-traffic")
def revision_update_deprecated(
    service: ResponseModelServiceInfo,
    number: List[int],
    weight: List[int],
    interactive: bool,
): 
    split(service=service, number=number, weight=weight, interactive=interactive)

@cli.command("split-traffic", cls=VesslCommand)
@service_name_option
@split_traffic_revs_option
@split_traffic_weights_option
def split_traffic(
        service: ResponseModelServiceInfo,
        number: List[int],
        weight: List[int],
):
    """
    Update revision's traffic weight.

    Example:
    
    $ vessl service split-traffic --service my-service -n 1 -w 100 
    $ vessl service split-traffic --service my-service -n 1 -w 60 -n 2 -w 40
    """
    split(service=service, number=number, weight=weight, interactive=True)


@cli.command("create-yaml", cls=VesslCommand, help="Create a service.yaml file with .vessl.model.lock file")
@vessl_argument("service_name", type=str)
@vessl_argument("message", type=str)
@vessl_option("-k", "--api-key", "use_api_key", type=bool, is_flag=True)
def service_create_yaml(
        service_name: str,
        message: str,
        use_api_key: bool = False,
): create_yaml(service_name=service_name, message=message, use_api_key=use_api_key)


@alias_cli.command("create-yaml", cls=VesslCommand, hidden=True)
@vessl_argument("service_name", type=str)
@vessl_argument("message", type=str)
@vessl_option("-k", "--api-key", "use_api_key", type=bool, is_flag=True)
def serve_create_yaml(
        service_name: str,
        message: str,
        use_api_key: bool = False,
): create_yaml(service_name=service_name, message=message, use_api_key=use_api_key)


@cli_gateway.command("show", cls=VesslCommand, hidden=True)
@service_name_option
@format_option
@deprecated(target_command="serve gateway show", new_command="service read")
def gateway_show(service: ResponseModelServiceInfo, format: str):
    """
    Show current status of the gateway of a service.
    """
    try:
        gateway = _read_service(service_name=service.name).gateway_config
    except VesslApiException as e:
        print(f"Failed to read gateway of service {service.name}: {e.message}")
        sys.exit(1)

    if format == "json":
        print(json.dumps(gateway.to_dict(), default=str))
    else:
        print_gateway(gateway)


@cli_revision.command("show", cls=VesslCommand, hidden=True)
@service_name_option
@revision_option
@format_option
@vessl_option("-o", "--output", "output", type=click.Path(), help="Output file path.")
@deprecated(target_command="serve revision show", new_command="service read")
def revision_show(service: ResponseModelServiceInfo, number: ResponseModelServiceRevision, format: str, output: str):
    """
    Show current status and information about a service revision.
    """
    revision = number
    result = ""
    if format == "text":
        print_revision(revision, verbose=True)
        return
    if format == "json":
        result = json.dumps(revision.to_dict(), default=str)
    if format == "yaml":
        result = revision.yaml_spec

    if output:
        with open(output, "w") as f:
            f.write(result)
    else:
        print_info(result)


@cli_revision.command("list", cls=VesslCommand)
@service_name_option
@format_option
@deprecated(target_command="serve revision list", new_command="service read")
def revision_list(service: ResponseModelServiceInfo, format: str):
    """
    List all revisions.
    """
    try:
        resp: ModelserviceModelServiceRevisionListResponse = vessl_api.model_service_revision_list_api(
        organization_name=vessl_api.organization.name,
        model_service_name=service.name,
        )
        revisions = resp.results
    except VesslApiException as e:
        print(f"Failed to list revisions of service {service.name}: {e.message}")
        sys.exit(1)

    if format == "json":
        print(json.dumps([r.to_dict() for r in revisions], default=str))
    if format == "yaml":
        print("\n\n".join([r.yaml_spec for r in revisions]))
    else:
        print(f"{len(revisions)} revision(s) found.\n")
        for i, revision in enumerate(revisions):
            if i > 0:
                print()
            print_revision(revision)


@cli.command("join", cls=VesslCommand, hidden=True)
def join():
    access_token = vessl_api.find_model_service_join_token()
    vessl_api.api_client.set_default_header("Authorization", f"Token {access_token}")
    resp = vessl_api.model_service_revision_workload_join_api()
    vessl_api.configure_jwt_access_token(resp.token)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmVolumeV2(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by_user_id': 'int',
        'created_by_workload_id': 'int',
        'created_dt': 'datetime',
        'edges': 'OrmVolumeV2Edges',
        'id': 'int',
        'immutable_slug': 'str',
        'is_auto_created': 'bool',
        'is_read_only': 'bool',
        'name': 'str',
        'origin_filesystem_path': 'str',
        'storage_id': 'int',
        'sub_path': 'str',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'created_by_user_id': 'created_by_user_id',
        'created_by_workload_id': 'created_by_workload_id',
        'created_dt': 'created_dt',
        'edges': 'edges',
        'id': 'id',
        'immutable_slug': 'immutable_slug',
        'is_auto_created': 'is_auto_created',
        'is_read_only': 'is_read_only',
        'name': 'name',
        'origin_filesystem_path': 'origin_filesystem_path',
        'storage_id': 'storage_id',
        'sub_path': 'sub_path',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, created_by_user_id=None, created_by_workload_id=None, created_dt=None, edges=None, id=None, immutable_slug=None, is_auto_created=None, is_read_only=None, name=None, origin_filesystem_path=None, storage_id=None, sub_path=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """OrmVolumeV2 - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by_user_id = None
        self._created_by_workload_id = None
        self._created_dt = None
        self._edges = None
        self._id = None
        self._immutable_slug = None
        self._is_auto_created = None
        self._is_read_only = None
        self._name = None
        self._origin_filesystem_path = None
        self._storage_id = None
        self._sub_path = None
        self._updated_dt = None
        self.discriminator = None

        self.created_by_user_id = created_by_user_id
        self.created_by_workload_id = created_by_workload_id
        if created_dt is not None:
            self.created_dt = created_dt
        if edges is not None:
            self.edges = edges
        if id is not None:
            self.id = id
        if immutable_slug is not None:
            self.immutable_slug = immutable_slug
        if is_auto_created is not None:
            self.is_auto_created = is_auto_created
        if is_read_only is not None:
            self.is_read_only = is_read_only
        if name is not None:
            self.name = name
        self.origin_filesystem_path = origin_filesystem_path
        if storage_id is not None:
            self.storage_id = storage_id
        if sub_path is not None:
            self.sub_path = sub_path
        if updated_dt is not None:
            self.updated_dt = updated_dt

    @property
    def created_by_user_id(self):
        """Gets the created_by_user_id of this OrmVolumeV2.  # noqa: E501


        :return: The created_by_user_id of this OrmVolumeV2.  # noqa: E501
        :rtype: int
        """
        return self._created_by_user_id

    @created_by_user_id.setter
    def created_by_user_id(self, created_by_user_id):
        """Sets the created_by_user_id of this OrmVolumeV2.


        :param created_by_user_id: The created_by_user_id of this OrmVolumeV2.  # noqa: E501
        :type created_by_user_id: int
        """

        self._created_by_user_id = created_by_user_id

    @property
    def created_by_workload_id(self):
        """Gets the created_by_workload_id of this OrmVolumeV2.  # noqa: E501


        :return: The created_by_workload_id of this OrmVolumeV2.  # noqa: E501
        :rtype: int
        """
        return self._created_by_workload_id

    @created_by_workload_id.setter
    def created_by_workload_id(self, created_by_workload_id):
        """Sets the created_by_workload_id of this OrmVolumeV2.


        :param created_by_workload_id: The created_by_workload_id of this OrmVolumeV2.  # noqa: E501
        :type created_by_workload_id: int
        """

        self._created_by_workload_id = created_by_workload_id

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmVolumeV2.  # noqa: E501


        :return: The created_dt of this OrmVolumeV2.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmVolumeV2.


        :param created_dt: The created_dt of this OrmVolumeV2.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def edges(self):
        """Gets the edges of this OrmVolumeV2.  # noqa: E501


        :return: The edges of this OrmVolumeV2.  # noqa: E501
        :rtype: OrmVolumeV2Edges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmVolumeV2.


        :param edges: The edges of this OrmVolumeV2.  # noqa: E501
        :type edges: OrmVolumeV2Edges
        """

        self._edges = edges

    @property
    def id(self):
        """Gets the id of this OrmVolumeV2.  # noqa: E501


        :return: The id of this OrmVolumeV2.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmVolumeV2.


        :param id: The id of this OrmVolumeV2.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this OrmVolumeV2.  # noqa: E501


        :return: The immutable_slug of this OrmVolumeV2.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this OrmVolumeV2.


        :param immutable_slug: The immutable_slug of this OrmVolumeV2.  # noqa: E501
        :type immutable_slug: str
        """

        self._immutable_slug = immutable_slug

    @property
    def is_auto_created(self):
        """Gets the is_auto_created of this OrmVolumeV2.  # noqa: E501


        :return: The is_auto_created of this OrmVolumeV2.  # noqa: E501
        :rtype: bool
        """
        return self._is_auto_created

    @is_auto_created.setter
    def is_auto_created(self, is_auto_created):
        """Sets the is_auto_created of this OrmVolumeV2.


        :param is_auto_created: The is_auto_created of this OrmVolumeV2.  # noqa: E501
        :type is_auto_created: bool
        """

        self._is_auto_created = is_auto_created

    @property
    def is_read_only(self):
        """Gets the is_read_only of this OrmVolumeV2.  # noqa: E501


        :return: The is_read_only of this OrmVolumeV2.  # noqa: E501
        :rtype: bool
        """
        return self._is_read_only

    @is_read_only.setter
    def is_read_only(self, is_read_only):
        """Sets the is_read_only of this OrmVolumeV2.


        :param is_read_only: The is_read_only of this OrmVolumeV2.  # noqa: E501
        :type is_read_only: bool
        """

        self._is_read_only = is_read_only

    @property
    def name(self):
        """Gets the name of this OrmVolumeV2.  # noqa: E501


        :return: The name of this OrmVolumeV2.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this OrmVolumeV2.


        :param name: The name of this OrmVolumeV2.  # noqa: E501
        :type name: str
        """

        self._name = name

    @property
    def origin_filesystem_path(self):
        """Gets the origin_filesystem_path of this OrmVolumeV2.  # noqa: E501


        :return: The origin_filesystem_path of this OrmVolumeV2.  # noqa: E501
        :rtype: str
        """
        return self._origin_filesystem_path

    @origin_filesystem_path.setter
    def origin_filesystem_path(self, origin_filesystem_path):
        """Sets the origin_filesystem_path of this OrmVolumeV2.


        :param origin_filesystem_path: The origin_filesystem_path of this OrmVolumeV2.  # noqa: E501
        :type origin_filesystem_path: str
        """

        self._origin_filesystem_path = origin_filesystem_path

    @property
    def storage_id(self):
        """Gets the storage_id of this OrmVolumeV2.  # noqa: E501


        :return: The storage_id of this OrmVolumeV2.  # noqa: E501
        :rtype: int
        """
        return self._storage_id

    @storage_id.setter
    def storage_id(self, storage_id):
        """Sets the storage_id of this OrmVolumeV2.


        :param storage_id: The storage_id of this OrmVolumeV2.  # noqa: E501
        :type storage_id: int
        """

        self._storage_id = storage_id

    @property
    def sub_path(self):
        """Gets the sub_path of this OrmVolumeV2.  # noqa: E501


        :return: The sub_path of this OrmVolumeV2.  # noqa: E501
        :rtype: str
        """
        return self._sub_path

    @sub_path.setter
    def sub_path(self, sub_path):
        """Sets the sub_path of this OrmVolumeV2.


        :param sub_path: The sub_path of this OrmVolumeV2.  # noqa: E501
        :type sub_path: str
        """

        self._sub_path = sub_path

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmVolumeV2.  # noqa: E501


        :return: The updated_dt of this OrmVolumeV2.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmVolumeV2.


        :param updated_dt: The updated_dt of this OrmVolumeV2.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmVolumeV2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmVolumeV2):
            return True

        return self.to_dict() != other.to_dict()



from enum import Enum


class ModelServiceType(Enum):
    PROVISIONED = 'provisioned'
    SERVERLESS = 'serverless'

class ClusterProvider(Enum):
    AWS = 'aws'
    GCP = 'gcp'
    OCI = 'oci'
    AZURE = 'azure'
    ON_PREMISE = 'on-premise'

class ClusterCustomEndpointType(Enum):
    NODE_PORT = 'nodeport'
    SUBDOMAIN = 'subdomain'


from concurrent.futures import wait
from time import time

from requests_futures.sessions import FuturesSession

from vessl.util.fmt import sizeof_fmt
from vessl.util.constant import PARALLEL_WORKERS


class FileTransmissionTracker:
    def __init__(self, total_count, total_size, action):
        self.start_time = time()
        self.total_count = total_count
        self.done_count = 0
        self.total_size = total_size
        self.done_size = 0
        self.last_printed_time = None
        self.action = action

    def increase_done_size(self, size: int):
        self.done_size += size

    def increase_done_count(self):
        self.done_count += 1

    def print_progress(self, force=False):
        if not self.last_printed_time:
            self.last_printed_time = time()

        elif force or time() - self.last_printed_time > 1:
            print(
                f"{sizeof_fmt(self.done_size)}/{sizeof_fmt(self.total_size)} "
                f"({int(self.done_size/self.total_size*100)}%) {self.action}ed."
            )
            self.last_printed_time = time()

    def print_result(self):
        self.print_progress(force=True)
        print(
            "Successfully {}ed {} out of {} file(s) in {} seconds".format(
                self.action,
                self.done_count,
                self.total_count,
                "%.2f" % (time() - self.start_time),
            )
        )


class FileTransmissionHandler:
    def __init__(self, future_func, action: str = "download", quiet: bool = False):
        self.session = FuturesSession(max_workers=PARALLEL_WORKERS)
        self.files = []
        self.total_count = 0
        self.total_size = 0
        self.quiet = quiet
        self.action = action
        self.future_func = future_func  # future_func(session, path, url, tracker) -> Future

    def add_file(self, full_path, url, size):
        self.total_count += 1
        self.total_size += size
        self.files.append((full_path, url))

    def run(self):
        tracker = None
        if not self.quiet:
            tracker = self._tracker()

        wait(
            [
                self.future_func(self.session, full_path, url, tracker)
                for full_path, url in self.files
            ]
        )
        if tracker:
            tracker.print_result()

    def _tracker(self):
        return FileTransmissionTracker(self.total_count, self.total_size, self.action)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class V1S3RefForm(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'bucket': 'str',
        'credential_id': 'int',
        'credential_name': 'str',
        'endpoint': 'str',
        'prefix': 'str'
    }

    attribute_map = {
        'bucket': 'bucket',
        'credential_id': 'credential_id',
        'credential_name': 'credential_name',
        'endpoint': 'endpoint',
        'prefix': 'prefix'
    }

    def __init__(self, bucket=None, credential_id=None, credential_name=None, endpoint=None, prefix=None, local_vars_configuration=None):  # noqa: E501
        """V1S3RefForm - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._bucket = None
        self._credential_id = None
        self._credential_name = None
        self._endpoint = None
        self._prefix = None
        self.discriminator = None

        self.bucket = bucket
        self.credential_id = credential_id
        self.credential_name = credential_name
        self.endpoint = endpoint
        self.prefix = prefix

    @property
    def bucket(self):
        """Gets the bucket of this V1S3RefForm.  # noqa: E501


        :return: The bucket of this V1S3RefForm.  # noqa: E501
        :rtype: str
        """
        return self._bucket

    @bucket.setter
    def bucket(self, bucket):
        """Sets the bucket of this V1S3RefForm.


        :param bucket: The bucket of this V1S3RefForm.  # noqa: E501
        :type bucket: str
        """
        if self.local_vars_configuration.client_side_validation and bucket is None:  # noqa: E501
            raise ValueError("Invalid value for `bucket`, must not be `None`")  # noqa: E501

        self._bucket = bucket

    @property
    def credential_id(self):
        """Gets the credential_id of this V1S3RefForm.  # noqa: E501


        :return: The credential_id of this V1S3RefForm.  # noqa: E501
        :rtype: int
        """
        return self._credential_id

    @credential_id.setter
    def credential_id(self, credential_id):
        """Sets the credential_id of this V1S3RefForm.


        :param credential_id: The credential_id of this V1S3RefForm.  # noqa: E501
        :type credential_id: int
        """

        self._credential_id = credential_id

    @property
    def credential_name(self):
        """Gets the credential_name of this V1S3RefForm.  # noqa: E501


        :return: The credential_name of this V1S3RefForm.  # noqa: E501
        :rtype: str
        """
        return self._credential_name

    @credential_name.setter
    def credential_name(self, credential_name):
        """Sets the credential_name of this V1S3RefForm.


        :param credential_name: The credential_name of this V1S3RefForm.  # noqa: E501
        :type credential_name: str
        """

        self._credential_name = credential_name

    @property
    def endpoint(self):
        """Gets the endpoint of this V1S3RefForm.  # noqa: E501


        :return: The endpoint of this V1S3RefForm.  # noqa: E501
        :rtype: str
        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint):
        """Sets the endpoint of this V1S3RefForm.


        :param endpoint: The endpoint of this V1S3RefForm.  # noqa: E501
        :type endpoint: str
        """

        self._endpoint = endpoint

    @property
    def prefix(self):
        """Gets the prefix of this V1S3RefForm.  # noqa: E501


        :return: The prefix of this V1S3RefForm.  # noqa: E501
        :rtype: str
        """
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        """Sets the prefix of this V1S3RefForm.


        :param prefix: The prefix of this V1S3RefForm.  # noqa: E501
        :type prefix: str
        """
        if self.local_vars_configuration.client_side_validation and prefix is None:  # noqa: E501
            raise ValueError("Invalid value for `prefix`, must not be `None`")  # noqa: E501

        self._prefix = prefix

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1S3RefForm):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1S3RefForm):
            return True

        return self.to_dict() != other.to_dict()



import base64
import os
import socket
from logging import Logger
from platform import uname
from time import sleep
from typing import TYPE_CHECKING, List, Optional, Tuple

import click

from vessl.openapi_client.models import (
    ClusterUpdateAPIInput,
    ResponseKernelClusterInfo,
    ResponseKernelClusterNodeInfo,
)
from vessl.openapi_client.models.response_kernel_resource_spec import (
    ResponseKernelResourceSpec,
)
from vessl import vessl_api
from vessl.kernel_cluster_install_command_builder import ClusterInstallCommandBuilder
from vessl.organization import _get_organization_name
from vessl.util import logger
from vessl.util.config import DEFAULT_VESSL_DIR
from vessl.util.constant import K0S_VERSION, VESSL_HELM_REPO
from vessl.util.exception import (
    ACCESS_TOKEN_NOT_FOUND_ERROR_CODE,
    ClusterNotFoundError,
    InvalidKernelClusterError,
    VesslApiException,
    VesslRuntimeException,
)

if TYPE_CHECKING:
    import docker


class CreateClusterParam:
    """Create cluster parameter for `vessl.create_cluster`.

    Attributes:
    - Required
        cluster_name(str): Cluster name.
        kubernetes_namespace(str): Kubernetes namespace to install VESSL agent.
            defaults to "vessl".
        extra_helm_values(list[str]): Helm values to pass to cluster install
            command. See https://github.com/vessl-ai/cluster-resources/blob/main/helm-chart/values.yaml
            for available Helm values.
        kubeconfig_path(str): Path to kubeconfig file.
        provider(str): Cluster provider. defaults to "on-premise"

    - Optional
        organization_name(str): Organization name. If not given, then use
            default organization.
        pod_resource_path(str): Path to pod resource file.
        enable_local_path_provisioner(bool): Whether to enable local path provisioner.
        enable_longhorn(bool): Whether to enable longhorn.
    """

    def __init__(
        self,
        cluster_name: str = "",
        kubeconfig_path: str = "",
        extra_helm_values: List[str] = [],
        provider: str = "on-premise",
        kubernetes_namespace: str = "vessl",
        organization_name: Optional[str] = None,
        pod_resource_path: str = "",
        enable_local_path_provisioner: bool = True,
        enable_longhorn: bool = False,
    ) -> None:
        self.cluster_name = cluster_name
        self.kubernetes_namespace = kubernetes_namespace
        self.extra_helm_values = extra_helm_values
        self.kubeconfig_path = kubeconfig_path
        self.provider = provider

        # optional
        self.organization_name: str = organization_name
        self.pod_resource_path = pod_resource_path
        self.enable_local_path_provisioner = enable_local_path_provisioner
        self.enable_longhorn = enable_longhorn


def create_cluster(
    param: CreateClusterParam,
) -> ResponseKernelClusterInfo:
    """Create a VESSL cluster by installing VESSL agent to given Kubernetes
    namespace. If you want to override the default organization, then pass
    `organization_name` to `param`.

    Args:
        param(CreateClusterParam): Create cluster parameter.

    Example:
        ```python
        vessl.install_cluster(
            param=vessl.CreateClusterParam(
                cluster_name="foo",
                ...
            ),
        )
        ```
    """
    agent_access_token = vessl_api.custom_cluster_key_api(
        organization_name=_get_organization_name(**{"organization_name": param.organization_name}),
    ).access_token

    logger.debug(f"cluster name: '{param.cluster_name}'")
    logger.debug(f"access token: '{agent_access_token}'")

    cluster_install_command = (
        ClusterInstallCommandBuilder()
        .with_kubeconfig(param.kubeconfig_path)
        .with_cluster_name(param.cluster_name)
        .with_provider_type(param.provider)
        .with_namespace(param.kubernetes_namespace)
        .with_access_token(agent_access_token)
        .with_local_path_provisioner_enabled(param.enable_local_path_provisioner)
        .with_longhorn_enabled(param.enable_longhorn)
        .with_helm_values(param.extra_helm_values)
        .with_pod_resource_path(param.pod_resource_path)
        .build()
    )

    stream = os.popen(cluster_install_command)
    output = stream.read()
    if stream.close() is not None:
        raise VesslRuntimeException(f"Failed to install VESSL agent: {output}")
    print(click.style(output, fg="blue"))

    logger.warn(
        click.style(
            "VESSL cluster agent installed. Waiting for the agent to be connected with VESSL...",
            fg="green",
        )
    )
    for attempts in range(18):
        sleep(10)
        try:
            return vessl_api.custom_cluster_connectivity_api(
                organization_name=_get_organization_name(
                    **{"organization_name": param.organization_name}
                ),
                agent_access_token=agent_access_token,
            )
        except VesslApiException as e:
            if e.code == "NotFound":
                continue
            raise e

    raise ClusterNotFoundError(
        "Timeout for checking agent connection; Please check agent log for more information."
    )


def add_and_update_helm_repo(logger: Logger = None):
    add_stream = os.popen(
        " ".join(
            [
                "helm",
                "repo",
                "add",
                "vessl",
                VESSL_HELM_REPO,
            ]
        )
    )
    repo_add_output = add_stream.read()
    if add_stream.close() != None:
        if logger:
            logger.debug("Failed to add helm repo")
        else:
            print(click.style("Failed to add helm repo", fg="red"))
        raise VesslRuntimeException(f"Failed to add helm repo: {repo_add_output}")

    update_stream = os.popen(" ".join(["helm", "repo", "update"]))
    repo_update_output = update_stream.read()

    if update_stream.close() != None:
        if logger:
            logger.debug("Failed to update helm repo")
        else:
            print(click.style("Failed to update helm repo", fg="red"))
        raise VesslRuntimeException(f"Failed to update helm repo: {repo_update_output}")

    # if no logger (ex. called from cli), print output
    if logger:
        logger.debug(repo_add_output)
        logger.debug(repo_update_output)
    else:
        print(click.style(repo_add_output, fg="blue"))
        print(click.style(repo_update_output, fg="blue"))


def _load_cluster_kubeconfig() -> Tuple[bool, Tuple[Optional[str], Optional[str]]]:
    """Find and load kubeconfig for VESSL cluster installation.

    Returns:
        True, if valid kubernetes config has loaded
        False, if no valid kubernetes config has found
    """
    from kubernetes import config
    from kubernetes.config import KUBE_CONFIG_DEFAULT_LOCATION

    cluster_kubeconfig = _get_cluster_kubeconfig()
    config_files = [cluster_kubeconfig] if cluster_kubeconfig else []
    config_files.append("/var/lib/k0s/pki/admin.conf")  # for k0s node as a worker
    config_files.append(os.path.expanduser(KUBE_CONFIG_DEFAULT_LOCATION))

    for config_file in config_files:
        try:
            config.load_kube_config(config_file)
            _, current_context = config.list_kube_config_contexts(config_file)
            logger.info(
                f"Using kubeconfig from file: {config_file} / context: {current_context['name']}"
            )
            return True, (config_file, current_context["name"])
        except config.config_exception.ConfigException as e:
            logger.warn(f"Cluster install: load kubeconfig from {config_file} failed: {repr(e)}")

    return False, (None, None)


def _install_k0s_in_docker() -> Tuple[str, str]:
    """Install k0s-in-Docker on local machine."""
    import docker

    logger.info(f"Installing k0s-in-docker ...")
    k0s_config = """apiVersion: k0s.k0sproject.io/v1beta1
kind: ClusterConfig
metadata:
  name: k0s
spec:
  api:
    extraArgs:
      service-node-port-range: 32639-32767
"""
    kwargs = {
        "name": "vessl-k0s",
        "environment": {"K0S_CONFIG": k0s_config},
        "hostname": f"vessl-k0s-{socket.gethostname()}",
        "ports": {
            "6443/tcp": 6443,
        },
        "privileged": True,
        "volumes": {
            "/var/lib/k0s": {
                "bind": "/var/lib/k0s",
                "mode": "rw",
            },
        },
    }
    for i in range(32639, 32768):
        kwargs["ports"][f"{i}/tcp"] = i

    # Add extra arguments for arm64 machines
    if uname().machine == "arm64":
        logger.debug("System is using CPU with arm64 architecture; adding additional kwargs")
        kwargs["environment"]["ETCD_UNSUPPORTED_ARCH"] = "arm64"

    client = docker.from_env()

    # Add extra arguments for docker container running on Docker Desktop
    if client.info().get("OperatingSystem", "") == "Docker Desktop":
        logger.debug("Docker is running on Docker Desktop; adding additional kwargs")
        kwargs["cgroupns"] = "host"
        kwargs["volumes"]["/sys/fs/cgroup"] = {
            "bind": "/sys/fs/cgroup",
            "mode": "rw",
        }

    # Run k0s container
    container = _start_k0s_container(kwargs)

    # Wait for the container to be ready
    _wait_for_k0s_ready(container)

    # Get kubeconfig, keep at /opt/vessl/cluster_kubeconfig.yaml for later
    logger.debug("Acquiring kubeconfig")
    kubeconfig_filename = _acquire_kubeconfig(container)

    current_context = _load_kubeconfig_to_config(kubeconfig_filename)
    logger.info(f"k0s-in-docker is ready! Saved kubeconfig at {kubeconfig_filename} for later use.")
    return kubeconfig_filename, current_context["name"]


def _load_kubeconfig_to_config(kubeconfig_filename):
    from kubernetes import config

    vessl_api.config_loader.cluster_kubeconfig = kubeconfig_filename
    config.load_kube_config(kubeconfig_filename)
    _, current_context = config.list_kube_config_contexts(kubeconfig_filename)
    return current_context


def _acquire_kubeconfig(container):
    res = container.exec_run("cat /var/lib/k0s/pki/admin.conf", stream=False, demux=True)
    output, _ = res.output
    kubeconfig = output.decode("utf-8")
    kubeconfig_filename = f"{DEFAULT_VESSL_DIR}/cluster_kubeconfig.yaml"
    with open(kubeconfig_filename, "w") as f:
        f.write(kubeconfig)
    return kubeconfig_filename


def _wait_for_k0s_ready(container):
    node_ready = False
    trial_limit = 10
    for trial in range(trial_limit):
        logger.info(f"Checking for the k0s-in-docker node to be ready... ({trial+1}/{trial_limit})")
        res = container.exec_run(
            "bash -c 'k0s kubectl get nodes --no-headers=true | grep -v NotReady | wc -l'",
            stream=False,
            demux=True,
        )
        output, err = res.output
        if output:
            logger.debug(output.decode("utf-8"))
            nodes = output.decode("utf-8").strip()
            if nodes == "1":
                node_ready = True
                break
        if err:
            logger.debug(err.decode("utf-8"))
        sleep(10)
    if not node_ready:
        container.kill()
        raise VesslRuntimeException("Timeout while waiting for k0s-in-docker node. Please retry.")


def _start_k0s_container(kwargs):
    import docker

    client = docker.from_env()
    container = client.containers.run(
        detach=True,
        image=f"docker.io/k0sproject/k0s:{K0S_VERSION}",
        command="k0s controller --single --no-taints --config /etc/k0s/config.yaml",
        **kwargs,
    )

    return container


def _get_running_k0s_container() -> "docker.models.containers.Container":
    import docker

    """Get k0s-in-Docker container."""
    client = docker.from_env()
    containers = client.containers.list(filters={"name": "vessl-k0s", "status": "running"})
    if len(containers) == 0:
        raise VesslRuntimeException("k0s-in-Docker container not found.")
    return containers[0]


def _check_existing_vessl_installation(
    namespace: str = "vessl", **kwargs
) -> ResponseKernelClusterInfo:
    """Check if there is existing vessl installation in current Kubernetes
    cluster."""
    from kubernetes import client as k8sclient

    # Find secret with name=vessl-agent, annotation=vessl
    try:
        v1 = k8sclient.CoreV1Api()
        vessl_secret: k8sclient.V1Secret = v1.read_namespaced_secret(
            name="vessl-agent", namespace=namespace
        )
    except k8sclient.ApiException as e:
        if e.status != 404:
            raise VesslRuntimeException("Error while communicating with kubernetes.")
        return None

    # VESSL agent token is found - there is existing cluster in current kubernetes cluster
    logger.warn("Existing VESSL cluster installation found! getting cluster information...")
    vessl_access_token = vessl_secret.data.get("access-token")
    if vessl_access_token is None:
        raise InvalidKernelClusterError(
            "Kubernetes secret for VESSL found, but secret does not have access-token",
            ACCESS_TOKEN_NOT_FOUND_ERROR_CODE,
        )
    vessl_access_token = base64.b64decode(vessl_access_token)
    logger.debug(f"access token: {vessl_access_token}")
    return vessl_api.custom_cluster_connectivity_api(
        organization_name=_get_organization_name(**kwargs),
        agent_access_token=vessl_access_token,
    )


def read_cluster(cluster_name: str, **kwargs) -> ResponseKernelClusterInfo:
    """Read cluster in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        cluster_name(str): Cluster name.

    Example:
        ```python
        vessl.read_cluster(
            cluster_name="seoul-cluster",
        )
        ```
    """
    kernel_clusters = list_clusters(**kwargs)
    kernel_clusters = {x.name: x for x in kernel_clusters}

    if cluster_name not in kernel_clusters:
        raise InvalidKernelClusterError(f"Cluster not found: {cluster_name}")
    return kernel_clusters[cluster_name]


def list_clusters(**kwargs) -> List[ResponseKernelClusterInfo]:
    """List clusters in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Example:
        ```python
        vessl.list_clusters()
        ```
    """
    return vessl_api.cluster_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).clusters


def delete_cluster(cluster_id: int, **kwargs) -> object:
    """Delete custom cluster in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.delete_cluster(
            cluster_id=1,
        )
        ```
    """
    return vessl_api.cluster_delete_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    )


def rename_cluster(cluster_id: int, new_cluster_name: str, **kwargs) -> ResponseKernelClusterInfo:
    """Rename custom cluster in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.
        new_cluster_name(str): Cluster name to change.

    Example:
        ```python
        vessl.rename_cluster(
            cluster_id=1,
            new_cluster_name="seoul-cluster-2",
        )
        ```
    """
    return vessl_api.cluster_update_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
        custom_cluster_update_api_input=ClusterUpdateAPIInput(
            name=new_cluster_name,
        ),
    )


def list_cluster_nodes(cluster_id: int, **kwargs) -> List[ResponseKernelClusterNodeInfo]:
    """List custom cluster nodes in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.list_cluster_nodes(
            cluster_id=1,
        )
        ```
    """
    return vessl_api.custom_cluster_node_list_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    ).nodes


def _get_cluster_kubeconfig(**kwargs) -> str:
    cluster_kubeconfig = kwargs.get("cluster_kubeconfig")
    if cluster_kubeconfig is not None:
        return cluster_kubeconfig
    if vessl_api.cluster_kubeconfig is not None:
        return vessl_api.cluster_kubeconfig

    return ""


def list_cluster_presets(cluster_id: int) -> List[ResponseKernelResourceSpec]:
    organization = _get_organization_name()
    presets = vessl_api.kernel_resource_spec_list_api(
        cluster_id=cluster_id,
        organization_name=organization,
    )

    return presets.results
